@@ -7,12 +7,13 @@
 # -----------------------------------------------------------------------------
 
 """
-Utilities
+Utilities to build a namespace view.
 """
 
 from __future__ import print_function
 
 from itertools import islice
+import inspect
 import re
 
 # Local imports
@@ -21,44 +22,41 @@ from spyder_kernels.py3compat import (NUMERIC_TYPES, INT_TYPES, TEXT_TYPES,
                                       is_type_text_string,
                                       is_binary_string, PY2,
                                       to_binary_string, iteritems)
+from spyder_kernels.utils.lazymodules import (
+    bs4, FakeObject, numpy as np, pandas as pd, PIL)
 
 
 #==============================================================================
-# FakeObject
+# Numpy support
 #==============================================================================
-class FakeObject(object):
-    """Fake class used in replacement of missing modules"""
-    pass
-
-
-#==============================================================================
-# Numpy arrays and numeric types support
-#==============================================================================
-try:
-    from numpy import (ndarray, array, matrix, recarray,
-                       int64, int32, int16, int8, uint64, uint32, uint16, uint8,
-                       float64, float32, float16, complex64, complex128, bool_)
-    from numpy.ma import MaskedArray
-    from numpy import savetxt as np_savetxt
-    from numpy import get_printoptions, set_printoptions
-except:
-    ndarray = array = matrix = recarray = MaskedArray = np_savetxt = \
-     int64 = int32 = int16 = int8 = uint64 = uint32 = uint16 = uint8 = \
-     float64 = float32 = float16 = complex64 = complex128 = bool_ = FakeObject
+def get_numeric_numpy_types():
+    return (np.int64, np.int32, np.int16, np.int8, np.uint64, np.uint32,
+            np.uint16, np.uint8, np.float64, np.float32, np.float16,
+            np.complex64, np.complex128, np.bool_)
 
 
 def get_numpy_dtype(obj):
-    """Return NumPy data type associated to obj
-    Return None if NumPy is not available
-    or if obj is not a NumPy array or scalar"""
-    if ndarray is not FakeObject:
-        # NumPy is available
-        import numpy as np
-        if isinstance(obj, np.generic) or isinstance(obj, np.ndarray):
-        # Numpy scalars all inherit from np.generic.
-        # Numpy arrays all inherit from np.ndarray.
-        # If we check that we are certain we have one of these
-        # types then we are less likely to generate an exception below.
+    """
+    Return Numpy data type associated to `obj`.
+
+    Return None if Numpy is not available, if we get errors or if `obj` is not
+    a Numpy array or scalar.
+    """
+    # Check if NumPy is available
+    if np.ndarray is not FakeObject:
+        # All Numpy scalars inherit from np.generic and all Numpy arrays
+        # inherit from np.ndarray. If we check that we are certain we have one
+        # of these types then we are less likely to generate an exception
+        # below.
+        # Note: The try/except is necessary to fix spyder-ide/spyder#19516.
+        try:
+            scalar_or_array = (
+                isinstance(obj, np.generic) or isinstance(obj, np.ndarray)
+            )
+        except Exception:
+            return
+
+        if scalar_or_array:
             try:
                 return obj.dtype.type
             except (AttributeError, RuntimeError):
@@ -67,33 +65,15 @@ def get_numpy_dtype(obj):
                 return
 
 
-#==============================================================================
-# Pandas support
-#==============================================================================
-try:
-    from pandas import DataFrame, DatetimeIndex, Series
-except:
-    DataFrame = DatetimeIndex = Series = FakeObject
-
-
-#==============================================================================
-# PIL Images support
-#==============================================================================
-try:
-    from spyder import pil_patch
-    Image = pil_patch.Image.Image
-except:
-    Image = FakeObject  # analysis:ignore
-
-
-#==============================================================================
-# BeautifulSoup support (see Issue 2448)
-#==============================================================================
-try:
-    import bs4
-    NavigableString = bs4.element.NavigableString
-except:
-    NavigableString = FakeObject  # analysis:ignore
+def get_numpy_type_string(value):
+    """Get the type of a Numpy object as a string."""
+    np_dtype = get_numpy_dtype(value)
+    if np_dtype is None or not hasattr(value, 'size'):
+        return 'Unknown'
+    elif value.size == 1:
+        return 'Scalar'
+    else:
+        return 'Array'
 
 
 #==============================================================================
@@ -112,18 +92,51 @@ def try_to_eval(value):
     except (NameError, SyntaxError, ImportError):
         return value
 
-    
+
 def get_size(item):
-    """Return size of an item of arbitrary type"""
-    if isinstance(item, (list, tuple, dict)):
-        return len(item)
-    elif isinstance(item, (ndarray, MaskedArray)):
-        return item.shape
-    elif isinstance(item, Image):
-        return item.size
-    if isinstance(item, (DataFrame, DatetimeIndex, Series)):
-        return item.shape
-    else:
+    """Return shape/size/len of an item of arbitrary type"""
+    try:
+        if (
+            hasattr(item, 'size') and hasattr(item.size, 'compute') or
+            hasattr(item, 'shape') and hasattr(item.shape, 'compute')
+        ):
+            # This is necessary to avoid an error when trying to
+            # get the size/shape of dask objects. We don't compute the
+            # size/shape since such operation could be expensive.
+            # Fixes spyder-ide/spyder#16844
+            return 1
+        elif (
+            hasattr(item, 'shape') and
+            isinstance(item.shape, (tuple, np.integer))
+        ):
+            try:
+                if item.shape:
+                    # This is needed since values could return as
+                    # `shape` an instance of a `tuple` subclass.
+                    # See spyder-ide/spyder#16348
+                    if isinstance(item.shape, tuple):
+                        return tuple(item.shape)
+                    return item.shape
+                else:
+                    # Scalar value
+                    return 1
+            except RecursionError:
+                # This is necessary to avoid an error when trying to
+                # get the shape of these objects.
+                # Fixes spyder-ide/spyder-kernels#217
+                return (-1, -1)
+        elif (hasattr(item, 'size') and
+                isinstance(item.size, (tuple, np.integer))):
+            try:
+                return item.size
+            except RecursionError:
+                return (-1, -1)
+        elif hasattr(item, '__len__'):
+            return len(item)
+        else:
+            return 1
+    except Exception:
+        # There is one item
         return 1
 
 
@@ -144,7 +157,6 @@ def get_object_attrs(obj):
 #==============================================================================
 import datetime
 
-
 try:
     from dateutil.parser import parse as dateparse
 except:
@@ -189,66 +201,48 @@ def str_to_timedelta(value):
 
 
 #==============================================================================
-# Background colors for supported types
+# Supported types
 #==============================================================================
-ARRAY_COLOR = "#00ff00"
-SCALAR_COLOR = "#0000ff"
-COLORS = {
-          bool:               "#ff00ff",
-          NUMERIC_TYPES:      SCALAR_COLOR,
-          list:               "#ffff00",
-          dict:               "#00ffff",
-          tuple:              "#c0c0c0",
-          TEXT_TYPES:         "#800000",
-          (ndarray,
-           MaskedArray,
-           matrix,
-           DataFrame,
-           Series,
-           DatetimeIndex):    ARRAY_COLOR,
-          Image:              "#008000",
-          datetime.date:      "#808000",
-          datetime.timedelta: "#808000",
-          }
-CUSTOM_TYPE_COLOR = "#7755aa"
-UNSUPPORTED_COLOR = "#ffffff"
-
-def get_color_name(value):
-    """Return color name depending on value type"""
-    if not is_known_type(value):
-        return CUSTOM_TYPE_COLOR
-    for typ, name in list(COLORS.items()):
-        if isinstance(value, typ):
-            return name
-    else:
-        np_dtype = get_numpy_dtype(value)
-        if np_dtype is None or not hasattr(value, 'size'):
-            return UNSUPPORTED_COLOR
-        elif value.size == 1:
-            return SCALAR_COLOR
-        else:
-            return ARRAY_COLOR
-
-
 def is_editable_type(value):
-    """Return True if data type is editable with a standard GUI-based editor,
-    like CollectionsEditor, ArrayEditor, QDateEdit or a simple QLineEdit"""
-    return get_color_name(value) not in (UNSUPPORTED_COLOR, CUSTOM_TYPE_COLOR)
+    """
+    Return True if data type is editable with a standard GUI-based editor,
+    like CollectionsEditor, ArrayEditor, QDateEdit or a simple QLineEdit.
+    """
+    if not is_known_type(value):
+        return False
+    else:
+        supported_types = [
+            'bool', 'int', 'long', 'float', 'complex', 'list', 'set', 'dict',
+            'tuple', 'str', 'unicode', 'NDArray', 'MaskedArray', 'Matrix',
+            'DataFrame', 'Series', 'PIL.Image.Image', 'datetime.date',
+            'datetime.timedelta'
+        ]
+
+        if (get_type_string(value) not in supported_types and
+                not isinstance(value, pd.Index)):
+            np_dtype = get_numpy_dtype(value)
+            if np_dtype is None or not hasattr(value, 'size'):
+                return False
+        return True
 
 
 #==============================================================================
 # Sorting
 #==============================================================================
-def sort_against(list1, list2, reverse=False):
+def sort_against(list1, list2, reverse=False, sort_key=None):
     """
     Arrange items of list1 in the same order as sorted(list2).
 
     In other words, apply to list1 the permutation which takes list2 
     to sorted(list2, reverse).
     """
+    if sort_key is None:
+        key = lambda x: x[0]
+    else:
+        key = lambda x: sort_key(x[0])
     try:
         return [item for _, item in 
-                sorted(zip(list2, list1), key=lambda x: x[0], reverse=reverse)]
+                sorted(zip(list2, list1), key=key, reverse=reverse)]
     except:
         return list1
 
@@ -267,18 +261,27 @@ def default_display(value, with_module=True):
     try:
         name = object_type.__name__
         module = object_type.__module__
+
+        # Classes correspond to new types
+        if name == 'type':
+            name = 'class'
+
         if with_module:
+            if name == 'module':
+                return value.__name__ + ' module'
+            if module == 'builtins':
+                return name + ' object'
             return name + ' object of ' + module + ' module'
-        else:
-            return name
-    except:
+        return name
+    except Exception:
         type_str = to_text_string(object_type)
         return type_str[1:-1]
 
 
 def collections_display(value, level):
-    """Display for collections (i.e. list, tuple and dict)."""
+    """Display for collections (i.e. list, set, tuple and dict)."""
     is_dict = isinstance(value, dict)
+    is_set = isinstance(value, set)
 
     # Get elements
     if is_dict:
@@ -289,10 +292,10 @@ def collections_display(value, level):
     # Truncate values
     truncate = False
     if level == 1 and len(value) > 10:
-        elements = islice(elements, 10) if is_dict else value[:10]
+        elements = islice(elements, 10) if is_dict or is_set else value[:10]
         truncate = True
     elif level == 2 and len(value) > 5:
-        elements = islice(elements, 5) if is_dict else value[:5]
+        elements = islice(elements, 5) if is_dict or is_set else value[:5]
         truncate = True
 
     # Get display of each element
@@ -315,6 +318,8 @@ def collections_display(value, level):
         display = '{' + display + '}'
     elif isinstance(value, list):
         display = '[' + display + ']'
+    elif isinstance(value, set):
+        display = '{' + display + '}'
     else:
         display = '(' + display + ')'
 
@@ -323,29 +328,26 @@ def collections_display(value, level):
 
 def value_to_display(value, minmax=False, level=0):
     """Convert value for display purpose"""
-    # To save current Numpy threshold
-    np_threshold = FakeObject
+    # To save current Numpy printoptions
+    np_printoptions = FakeObject
+    numeric_numpy_types = get_numeric_numpy_types()
 
     try:
-        numeric_numpy_types = (int64, int32, int16, int8,
-                               uint64, uint32, uint16, uint8,
-                               float64, float32, float16,
-                               complex128, complex64, bool_)
-        if ndarray is not FakeObject:
-            # Save threshold
-            np_threshold = get_printoptions().get('threshold')
+        if np.ndarray is not FakeObject:
+            # Save printoptions
+            np_printoptions = np.get_printoptions()
             # Set max number of elements to show for Numpy arrays
             # in our display
-            set_printoptions(threshold=10)
-        if isinstance(value, recarray):
+            np.set_printoptions(threshold=10)
+        if isinstance(value, np.recarray):
             if level == 0:
                 fields = value.names
                 display = 'Field names: ' + ', '.join(fields)
             else:
                 display = 'Recarray'
-        elif isinstance(value, MaskedArray):
+        elif isinstance(value, np.ma.MaskedArray):
             display = 'Masked array'
-        elif isinstance(value, ndarray):
+        elif isinstance(value, np.ndarray):
             if level == 0:
                 if minmax:
                     try:
@@ -361,14 +363,14 @@ def value_to_display(value, minmax=False, level=0):
                     display = default_display(value)
             else:
                 display = 'Numpy array'
-        elif any([type(value) == t for t in [list, tuple, dict]]):
+        elif any([type(value) == t for t in [list, set, tuple, dict]]):
             display = collections_display(value, level+1)
-        elif isinstance(value, Image):
+        elif isinstance(value, PIL.Image.Image):
             if level == 0:
                 display = '%s  Mode: %s' % (address(value), value.mode)
             else:
                 display = 'Image'
-        elif isinstance(value, DataFrame):
+        elif isinstance(value, pd.DataFrame):
             if level == 0:
                 cols = value.columns
                 if PY2 and len(cols) > 0:
@@ -387,19 +389,19 @@ def value_to_display(value, minmax=False, level=0):
                 display = 'Column names: ' + ', '.join(list(cols))
             else:
                 display = 'Dataframe'
-        elif isinstance(value, NavigableString):
+        elif isinstance(value, bs4.element.NavigableString):
             # Fixes Issue 2448
             display = to_text_string(value)
             if level > 0:
                 display = u"'" + display + u"'"
-        elif isinstance(value, DatetimeIndex):
+        elif isinstance(value, pd.Index):
             if level == 0:
                 try:
                     display = value._summary()
                 except AttributeError:
                     display = value.summary()
             else:
-                display = 'DatetimeIndex'
+                display = 'Index'
         elif is_binary_string(value):
             # We don't apply this to classes that extend string types
             # See issue 5636
@@ -435,7 +437,7 @@ def value_to_display(value, minmax=False, level=0):
                 display = default_display(value)
             else:
                 display = default_display(value, with_module=False)
-    except:
+    except Exception:
         display = default_display(value)
 
     # Truncate display at 70 chars to avoid freezing Spyder
@@ -447,9 +449,9 @@ def value_to_display(value, minmax=False, level=0):
             ellipses = u' ...'
         display = display[:70].rstrip() + ellipses
 
-    # Restore Numpy threshold
-    if np_threshold is not FakeObject:
-        set_printoptions(threshold=np_threshold)
+    # Restore Numpy printoptions
+    if np_printoptions is not FakeObject:
+        np.set_printoptions(**np_printoptions)
 
     return display
 
@@ -508,36 +510,56 @@ def display_to_value(value, default_value, ignore_errors=True):
 # =============================================================================
 def get_type_string(item):
     """Return type string of an object."""
-    if isinstance(item, DataFrame):
-        return "DataFrame"
-    if isinstance(item, DatetimeIndex):
-        return "DatetimeIndex"
-    if isinstance(item, Series):
-        return "Series"
+    # The try/except is necessary to fix spyder-ide/spyder#19516.
+    try:
+        # Numpy objects (don't change the order!)
+        if isinstance(item, np.ma.MaskedArray):
+            return "MaskedArray"
+        if isinstance(item, np.matrix):
+            return "Matrix"
+        if isinstance(item, np.ndarray):
+            return "NDArray"
+
+        # Pandas objects
+        if isinstance(item, pd.DataFrame):
+            return "DataFrame"
+        if isinstance(item, pd.Index):
+            return type(item).__name__
+        if isinstance(item, pd.Series):
+            return "Series"
+    except Exception:
+        pass
+
     found = re.findall(r"<(?:type|class) '(\S*)'>",
                        to_text_string(type(item)))
     if found:
+        if found[0] == 'type':
+            return 'class'
         return found[0]
-    
+    else:
+        return 'Unknown'
+
 
 def is_known_type(item):
     """Return True if object has a known type"""
     # Unfortunately, the masked array case is specific
-    return isinstance(item, MaskedArray) or get_type_string(item) is not None
+    return (isinstance(item, np.ma.MaskedArray) or
+            get_type_string(item) != 'Unknown')
 
 
 def get_human_readable_type(item):
     """Return human-readable type string of an item"""
-    if isinstance(item, (ndarray, MaskedArray)):
-        return item.dtype.name
-    elif isinstance(item, Image):
-        return "Image"
-    else:
-        text = get_type_string(item)
-        if text is None:
-            text = to_text_string('unknown')
+    # The try/except is necessary to fix spyder-ide/spyder#19516.
+    try:
+        if isinstance(item, (np.ndarray, np.ma.MaskedArray)):
+            return u'Array of ' + item.dtype.name
+        elif isinstance(item, PIL.Image.Image):
+            return "Image"
         else:
+            text = get_type_string(item)
             return text[text.find('.')+1:]
+    except Exception:
+        return 'Unknown'
 
 
 #==============================================================================
@@ -545,11 +567,13 @@ def get_human_readable_type(item):
 # CollectionsEditor)
 #==============================================================================
 def is_supported(value, check_all=False, filters=None, iterate=False):
-    """Return True if the value is supported, False otherwise"""
+    """Return True if value is supported, False otherwise."""
     assert filters is not None
     if value is None:
         return True
-    if not is_editable_type(value):
+    if is_callable_or_module(value):
+        return True
+    elif not is_editable_type(value):
         return False
     elif not isinstance(value, filters):
         return False
@@ -573,21 +597,34 @@ def is_supported(value, check_all=False, filters=None, iterate=False):
     return True
 
 
+def is_callable_or_module(value):
+    """Return True if value is a callable or module, False otherwise."""
+    try:
+        callable_or_module = callable(value) or inspect.ismodule(value)
+    except Exception:
+        callable_or_module = False
+    return callable_or_module
+
+
 def globalsfilter(input_dict, check_all=False, filters=None,
                   exclude_private=None, exclude_capitalized=None,
                   exclude_uppercase=None, exclude_unsupported=None,
-                  excluded_names=None):
-    """Keep only objects that can be pickled"""
+                  excluded_names=None, exclude_callables_and_modules=None):
+    """Keep objects in namespace view according to different criteria."""
     output_dict = {}
+    _is_string = is_type_text_string
+
     for key, value in list(input_dict.items()):
-        excluded = (exclude_private and key.startswith('_')) or \
-                   (exclude_capitalized and key[0].isupper()) or \
-                   (exclude_uppercase and key.isupper()
-                    and len(key) > 1 and not key[1:].isdigit()) or \
-                   (key in excluded_names) or \
-                   (exclude_unsupported and \
-                    not is_supported(value, check_all=check_all,
-                                     filters=filters))
+        excluded = (
+            (exclude_private and _is_string(key) and key.startswith('_')) or
+            (exclude_capitalized and _is_string(key) and key[0].isupper()) or
+            (exclude_uppercase and _is_string(key) and key.isupper() and
+             len(key) > 1 and not key[1:].isdigit()) or
+            (key in excluded_names) or
+            (exclude_callables_and_modules and is_callable_or_module(value)) or
+            (exclude_unsupported and
+             not is_supported(value, check_all=check_all, filters=filters))
+        )
         if not excluded:
             output_dict[key] = value
     return output_dict
@@ -598,7 +635,8 @@ def globalsfilter(input_dict, check_all=False, filters=None,
 #==============================================================================
 REMOTE_SETTINGS = ('check_all', 'exclude_private', 'exclude_uppercase',
                    'exclude_capitalized', 'exclude_unsupported',
-                   'excluded_names', 'minmax')
+                   'excluded_names', 'minmax', 'show_callable_attributes',
+                   'show_special_attributes', 'exclude_callables_and_modules')
 
 
 def get_supported_types():
@@ -611,21 +649,21 @@ def get_supported_types():
     in spyder-docs
     """
     from datetime import date, timedelta
-    editable_types = [int, float, complex, list, dict, tuple, date, timedelta
-                      ] + list(TEXT_TYPES) + list(INT_TYPES)
+    editable_types = [int, float, complex, list, set, dict, tuple, date,
+                      timedelta] + list(TEXT_TYPES) + list(INT_TYPES)
     try:
         from numpy import ndarray, matrix, generic
         editable_types += [ndarray, matrix, generic]
     except:
         pass
     try:
-        from pandas import DataFrame, Series, DatetimeIndex
-        editable_types += [DataFrame, Series, DatetimeIndex]
+        from pandas import DataFrame, Series, Index
+        editable_types += [DataFrame, Series, Index]
     except:
         pass
     picklable_types = editable_types[:]
     try:
-        from spyder.pil_patch import Image
+        from PIL import Image
         editable_types.append(Image.Image)
     except:
         pass
@@ -642,16 +680,19 @@ def get_remote_data(data, settings, mode, more_excluded_names=None):
     """
     supported_types = get_supported_types()
     assert mode in list(supported_types.keys())
-    excluded_names = settings['excluded_names']
+    excluded_names = list(settings['excluded_names'])
     if more_excluded_names is not None:
         excluded_names += more_excluded_names
-    return globalsfilter(data, check_all=settings['check_all'],
-                         filters=tuple(supported_types[mode]),
-                         exclude_private=settings['exclude_private'],
-                         exclude_uppercase=settings['exclude_uppercase'],
-                         exclude_capitalized=settings['exclude_capitalized'],
-                         exclude_unsupported=settings['exclude_unsupported'],
-                         excluded_names=excluded_names)
+    return globalsfilter(
+        data,
+        check_all=settings['check_all'],
+        filters=tuple(supported_types[mode]),
+        exclude_private=settings['exclude_private'],
+        exclude_uppercase=settings['exclude_uppercase'],
+        exclude_capitalized=settings['exclude_capitalized'],
+        exclude_unsupported=settings['exclude_unsupported'],
+        exclude_callables_and_modules=settings['exclude_callables_and_modules'],
+        excluded_names=excluded_names)
 
 
 def make_remote_view(data, settings, more_excluded_names=None):
@@ -664,8 +705,12 @@ def make_remote_view(data, settings, more_excluded_names=None):
     remote = {}
     for key, value in list(data.items()):
         view = value_to_display(value, minmax=settings['minmax'])
-        remote[key] = {'type':  get_human_readable_type(value),
-                       'size':  get_size(value),
-                       'color': get_color_name(value),
-                       'view':  view}
+        remote[key] = {
+            'type':  get_human_readable_type(value),
+            'size':  get_size(value),
+            'view':  view,
+            'python_type': get_type_string(value),
+            'numpy_type': get_numpy_type_string(value)
+        }
+
     return remote
